package gs.smolban.db.doobie

import cats.data.EitherT
import cats.syntax.all.*
import doobie.*
import doobie.implicits.*
import gs.smolban.db.DbError
import gs.smolban.db.TicketDb
import gs.smolban.db.doobie.DoobieTypes.*
import gs.smolban.model.CreatedAt
import gs.smolban.model.Group
import gs.smolban.model.Tag
import gs.smolban.model.Ticket
import java.time.Instant

final class DoobieTicketDb() extends TicketDb[ConnectionIO]:
  import DoobieTicketDb.Sql

  /** @inheritdoc
    */
  override def createTicket(ticket: Ticket)
    : EitherT[ConnectionIO, DbError, Ticket] = ???

  /** @inheritdoc
    */
  override def readTicket(ref: Ticket.Reference): ConnectionIO[Option[Ticket]] =
    Sql.readTicket(ref.number, ref.group).option.flatMap {
      case None => None.pure[ConnectionIO]
      case Some(ticketContents) =>
        for
          tags <- Sql.readTagsForTicket(ticketContents.id).stream.compile.toList
          history <- Sql
            .readHistoryForTicket(ticketContents.id)
            .stream
            .compile
            .toList
        yield Some(
          Ticket(
            number = ref.number,
            group = ref.group,
            createdAt = ticketContents.createdAt,
            title = ticketContents.title,
            description = ticketContents.description,
            tags = tags,
            status = ticketContents.status,
            statusHistory = history
          )
        )
    }

  /** @inheritdoc
    */
  override def updateTicket(
    ref: Ticket.Reference,
    newValue: Ticket
  ): ConnectionIO[Option[Ticket]] = ???

  /** @inheritdoc
    */
  override def deleteTicket(ref: Ticket.Reference): ConnectionIO[Boolean] = ???

object DoobieTicketDb:

  private object Sql:

    case class TicketContents(
      id: Long,
      createdAt: CreatedAt,
      title: String,
      description: String,
      status: Ticket.Status
    )

    def readTicket(
      ticketNumber: Ticket.Number,
      groupId: Group.Id
    ): Query0[TicketContents] = sql"""
      SELECT id, created_at, created_by, title, description, status, assignee
      FROM tickets
      WHERE ticket_number = $ticketNumber AND group_id = $groupId
      """.query[TicketContents]

    def readTagsForTicket(
      ticketId: Long
    ): Query0[Tag] = sql"""
    SELECT id, tag_value, created_at
    FROM tags
    WHERE ticket_id = $ticketId
    """.query[Tag]

    def readHistoryForTicket(
      ticketId: Long
    ): Query0[(Ticket.Status, Instant)] = sql"""
    SELECT status, set_at
    FROM ticket_history
    WHERE ticket_id = $ticketId
    """.query[(Ticket.Status, Instant)]

  end Sql

end DoobieTicketDb
